import os
import sys

import numpy as np

import pyemma
from paths import missing

print('pyemma path:', pyemma.__path__)

sys.path.append('/home/mi/marscher/workspace/pyemma/feature_sel/analysis/')
sys.path.append('/home/feature_sel/analysis/')
from pyemma._ext.sklearn.parameter_search import ParameterGrid

test_systems = [
   '1FME',
   '2F4K',
   '2JOF',
   '2WAV',
   'A3D',
   'CLN025',
   'GTT',
   'lambda',
   'NuG2',
   'PRB',
   'UVF',
   'NTL9',
]

features = (
    #'xyz',
            #'dist_ca', # TODO: this is the same feature as res_mindist basically.
            # 'flex_torsions',
            # #'shrake_ruply',
            # 'res_mindist',
            # 'res_mindist_d1',
            # 'res_mindist_d2',
            # 'res_mindist_expd',
            # 'res_mindist_c_0.4',
            # 'res_mindist_c_0.5',
            # 'res_mindist_c_0.6',
            # 'res_mindist_c_0.8',
            # 'res_mindist_c_1.0',
    'res_mindist_log',
            )


import argparse
parser = argparse.ArgumentParser()
parser.add_argument('--output', default='/group/ag_cmb/marscher/feature_sel/')
parser.add_argument('--lag', type=int, default=500) # dt=200ps, lag=500ns
parser.add_argument('id_', type=int)
args = parser.parse_args()

output_path = args.output
assert os.path.exists(output_path)

# prefix test sys and feature to force sorting group these.
grid = ParameterGrid([{'0_test_system': test_systems,
                       'lag': [args.lag],  # dt=200ps, lag=500ns
                       '1_feature': list(features),
                       #'k': [3,4,5,6,7,8,9,10],
                       }])

print('grid len:', len(grid))
#for i, e in enumerate(grid):
#    print(i, e)

def __myrepr__(x):
    s = ",".join([str(x_) for x_ in x])
    return s

#print("missing:\n%s" % __myrepr__(missing(grid=grid, output_path=output_path)))


#@app.task
def run(id_):
    print('job number: ', id_)
    params = grid[id_]
    feature = params['1_feature']
    test_system = params['0_test_system']
    lag = params['lag']
    pyemma.config.coordinates_check_output = True
    import estimate as e
    from paths import get_output_file_name
    cov = None
    try:
        for k in (5, 10, None):  #[1,2,3,4,5,6,7,8,9,10]:
            grid.param_grid[0]['k'] = [k]
            #print(grid.param_grid)
            fname = get_output_file_name(grid, id_)
            fname = os.path.join(output_path, fname)
            # TODO: exclude k param from cov file name to avoid recomp.
            fname_cov = os.path.join(output_path, get_output_file_name(grid, id_, include_k=False) + '_covs.h5')
            print('output file: %s' % fname)
            if os.path.exists(fname):
                print("results file %s already exists. Skipping" % fname)
                continue
            parameters = np.array({'lag': lag,
                                   'splitter': 'kfold',
                                   'mode': 'sliding',
                                   'test_system': test_system,
                                   'scoring_method': 'vamp2',
                                   'feature': feature,
                                   'k': k
                                   })

            if not os.path.exists(fname_cov):
                print('estimating covs')
                cov = e.estimate_covs(test_system=test_system, feature=feature, lag=lag, n_covs=100)
                cov.save(fname_cov)
            elif cov is None:
                print('loading covs from', fname_cov)
                cov = pyemma.load(fname_cov)
            assert cov is not None
            e.score_and_save(cov,
                             splitter='kfold',
                             fname=fname,
                             k=k,
                             fixed_seed=False,
                             scoring_method='VAMP2',
                             parameters=parameters, n_splits=50)

    except BaseException as e:
        print('bad: %s. failed id: %s' %(e, id_))
        import traceback
        traceback.print_exc()
    else:
        print('successful')


if __name__ == '__main__':
    run(args.id_)



#!C:\bld\pyemma_1550861281410\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'pyEMMA==2.5.5','console_scripts','pyemma_list_models'
__requires__ = 'pyEMMA==2.5.5'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('pyEMMA==2.5.5', 'console_scripts', 'pyemma_list_models')()
    )




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Created on 28.10.2013

@author: marscher
'''
from numpy.testing import assert_allclose as assert_allclose_np

__all__ = ['assert_allclose',
           ]


def assert_allclose(actual, desired, rtol=1.e-5, atol=1.e-8,
                    err_msg='', verbose=True):
    r"""wrapper for numpy.testing.allclose with default tolerances of
    numpy.allclose. Needed since testing method has different values."""
    return assert_allclose_np(actual, desired, rtol=rtol, atol=atol,
                              err_msg=err_msg, verbose=verbose)


def _hash_numpy_array(x):
    import numpy as np
    x = np.ascontiguousarray(x)
    import hashlib
    old = x.flags.writeable
    try:
        x.flags.writeable = False
        v = hashlib.sha1(x)
    finally:
        x.flags.writeable = old
    hash_value = hash((x.shape, x.strides, v.digest()))

    return hash_value



# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
Created on 09.04.2015

@author: marscher
'''

import numpy as np

import unittest
import tempfile
import os

from pyemma.coordinates.data.py_csv_reader import PyCSVReader as CSVReader
import shutil


class TestCSVReader(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.dir = tempfile.mkdtemp(prefix='pyemma_filereader')
        cls.nt = 300
        cls.nd = 4
        cls.data = np.arange(cls.nt * cls.nd).reshape(cls.nt, cls.nd)
        cls.filename1 = os.path.join(cls.dir, "data.dat")
        cls.filename2 = os.path.join(cls.dir, "data2.dat")
        np.savetxt(cls.filename1, cls.data)
        np.savetxt(cls.filename2, cls.data)
    
        cls.file_with_header = tempfile.mktemp(suffix=".dat", dir=cls.dir)
        cls.file_with_header2 = tempfile.mktemp(suffix=".dat", dir=cls.dir)
    
        np.savetxt(cls.file_with_header, cls.data, header="x y z")
        np.savetxt(cls.file_with_header2, cls.data, header="x y z")
        return cls

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.dir, ignore_errors=True)

    def test_read_1file(self):
        reader = CSVReader(self.filename1, chunksize=30)

        self.assertEqual(reader.number_of_trajectories(), 1)
        self.assertEqual(reader.dimension(), self.nd)
        self.assertEqual(reader.n_frames_total(), self.nt)

        output = reader.get_output()

        np.testing.assert_almost_equal(output[0], self.data)

    def test_read_1file_oneline(self):
        tiny = np.array([1, 2, 3])
        with tempfile.NamedTemporaryFile(mode='wb', suffix='.dat', delete=False) as f:
            np.savetxt(f, tiny)
            f.close()
            reader = CSVReader(f.name, delimiters=" ")
            np.testing.assert_equal(reader.get_output()[0], np.atleast_2d(tiny).T)

    def test_read_1file_with_header_2lines(self):
        data = np.array([1,2,3])
        with tempfile.NamedTemporaryFile(delete=False) as f:
            np.savetxt(f.name, data, header='x')
            f.close()
            out = CSVReader(f.name, delimiters=' ').get_output()[0]
            np.testing.assert_equal(out, np.atleast_2d(data).T)

    def test_read_1file_with_header(self):
        reader = CSVReader(self.file_with_header)
        self.assertEqual(reader.number_of_trajectories(), 1)
        self.assertEqual(reader.dimension(), self.nd)
        self.assertEqual(reader.n_frames_total(), self.nt)

        output = reader.get_output()

        np.testing.assert_almost_equal(output[0], self.data)

    def test_read_2file_with_header(self):
        reader = CSVReader([self.file_with_header, self.file_with_header2])
        self.assertEqual(reader.number_of_trajectories(), 2)
        self.assertEqual(reader.dimension(), self.nd)
        self.assertEqual(reader.n_frames_total(), self.nt*2)

        output = reader.get_output()

        np.testing.assert_almost_equal(output[0], self.data)
        np.testing.assert_almost_equal(output[1], self.data)

    def test_read_with_skipping_first_few_couple_lines(self):
        for skip in [0, 3, 13]:
            # FIXME: opening the same file twice is not being liked by py27
            r1 = CSVReader(self.filename1, chunksize=30)
            out_with_skip = r1.get_output(skip=skip)[0]
            assert len(out_with_skip) == len(self.data[skip:])
            r2 = CSVReader(self.filename1, chunksize=30)
            self.maxDiff=None
            #self.assertDictEqual(r1.__dict__, r2.__dict__)
            out = r2.get_output()[0]
            np.testing.assert_almost_equal(out_with_skip, out[skip::],
                                           err_msg="The first %s rows were skipped, but that did not "
                                                   "match the rows with skip=0 and sliced by [%s::]" % (skip, skip))

    def test_read_with_skipping_first_few_couple_lines_multiple_trajectoryfiles(self):
        for skip in [0, 3, 13]:
            r1 = CSVReader([self.filename1, self.filename2])
            out_with_skip = r1.get_output(skip=skip)
            r2 = CSVReader([self.filename1, self.filename2])
            out = r2.get_output()
            np.testing.assert_almost_equal(out_with_skip[0], out[0][skip::],
                                           err_msg="The first %s rows of the first file were skipped, but that did not "
                                                   "match the rows with skip=0 and sliced by [%s::]" % (skip, skip))
            np.testing.assert_almost_equal(out_with_skip[1], out[1][skip::],
                                           err_msg="The first %s rows of the second file were skipped, but that did not"
                                                   " match the rows with skip=0 and sliced by [%s::]" % (skip, skip))

    def test_read_lagged_small_chunks(self):
        lag = 200
        reader = CSVReader(self.filename1, chunksize=30)

        lagged_data = self.data[lag:]

        lagged_chunks = []
        it = reader.iterator(lag=lag)
        with it:
            for _, _, Y in it:
                assert len(Y) > 0
                lagged_chunks.append(Y)

        lagged_chunks = np.vstack(lagged_chunks)

        np.testing.assert_almost_equal(lagged_chunks, lagged_data)

    def test_with_kwargs(self):
        args = {'header': 27}

        reader = CSVReader(self.filename1, **args)

        output = reader.get_output()
        np.testing.assert_almost_equal(output[0], self.data)

    def test_with_multiple_files(self):
        files = [self.filename1, self.file_with_header]
        reader = CSVReader(files)

        self.assertEqual(reader.number_of_trajectories(), len(files))

    def test_with_stride(self):
        reader = CSVReader(self.filename1)

        for s in [2, 3, 7, 10]:
            output = reader.get_output(stride=s)[0]
            np.testing.assert_almost_equal(output, self.data[::s], err_msg="stride=%s"%s)

    def test_with_binary_written_file(self):
        data = np.arange(9).reshape(3, 3)
        with tempfile.NamedTemporaryFile('w+b', delete=False) as tmp:
            np.savetxt(tmp.name, data)
            tmp.close()
            out = CSVReader(tmp.name).get_output()[0]
        np.testing.assert_allclose(out, data)

    def test_with_lag(self):
        reader = CSVReader(self.filename1)

        for t in [23, 7, 59]:
            chunks = []
            it = reader.iterator(stride=1, lag=t)
            with it:
                for _, _, Y in it:
                    chunks.append(Y)
            chunks = np.vstack(chunks)
            np.testing.assert_almost_equal(chunks, self.data[t:])

    def test_with_stride_and_lag(self):
        reader = CSVReader(self.filename1)

        for s in [2, 3, 7, 10]:
            for t in [1, 23, 7, 59]:
                chunks = []
                chunks_lag = []
                it = reader.iterator(stride=s, lag=t)
                with it:
                    for _, X, Y in it:
                        chunks.append(X)
                        chunks_lag.append(Y)
                chunks = np.vstack(chunks)
                chunks_lag = np.vstack(chunks_lag)
                actual_lagged = self.data[t::s]
                np.testing.assert_almost_equal(chunks, self.data[::s][0:len(actual_lagged)],
                                               err_msg="output is not equal for"
                                                       " lag %i and stride %i" % (t, s)
                                               )
                np.testing.assert_almost_equal(chunks_lag, self.data[t::s],
                                               err_msg="output is not equal for"
                                                       " lag %i and stride %i" % (t, s))

    def test_with_stride_and_lag_with_header(self):
        reader = CSVReader(self.file_with_header)

        for s in [2, 3, 7, 10]:
            for t in [1, 23, 7, 59]:
                chunks = []
                chunks_lag = []
                it = reader.iterator(stride=s, lag=t)
                with it:
                    for _, X, Y in it:
                        chunks.append(X)
                        chunks_lag.append(Y)
                chunks = np.vstack(chunks)
                chunks_lag = np.vstack(chunks_lag)
                actual_lagged = self.data[t::s]
                np.testing.assert_almost_equal(chunks, self.data[::s][0:len(actual_lagged)])
                np.testing.assert_almost_equal(chunks_lag, self.data[t::s],
                                               err_msg="output is not equal for"
                                                       " lag %i and stride %i" % (t, s))

    def test_compare_readline(self):
        data = np.arange(99*3).reshape(-1, 3)
        with tempfile.NamedTemporaryFile(delete=False) as f:
            fn = f.name
            np.savetxt(fn, data)
            # calc offsets
            reader = CSVReader(fn)
            assert reader.dimension() == 3
            trajinfo = reader._get_traj_info(fn)
            offset = [0]
            with open(fn, CSVReader.DEFAULT_OPEN_MODE) as fh2:
                while fh2.readline():
                    offset.append(fh2.tell())
                fh2.seek(0)
                np.testing.assert_equal(trajinfo.offsets, offset)
                for ii, off in enumerate(trajinfo.offsets):
                    fh2.seek(off)
                    line = fh2.readline()
                    fh2.seek(offset[ii])
                    line2 = fh2.readline()

                    self.assertEqual(line, line2, "differs at offset %i (%s != %s)" % (ii, off, offset[ii]))

    def test_use_cols(self):
        reader = CSVReader(self.filename1)
        cols = (0, 2)
        with reader.iterator(chunk=0, cols=cols, return_trajindex=False) as it:
            for x in it:
                np.testing.assert_equal(x, self.data[:, cols])

    def test_newline_at_eof(self):
        x = "1 2 3\n4 5 6\n\n"
        desired = np.fromstring(x, sep=" ", dtype=np.float32).reshape(-1, 3)
        assert len(desired) == 2
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as f:
            f.write(x)
            f.close()
            reader = CSVReader(f.name)
            result = reader.get_output()[0]
            np.testing.assert_allclose(result, desired)

    def test_newline_at_eof_with_header(self):
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as f:
            f.write("#x y z\n1 2 3\n4 5 6\n\n")
            f.close()
            desired = np.genfromtxt(f.name, dtype=np.float32).reshape(-1, 3)
            reader = CSVReader(f.name)
            result = reader.get_output()[0]
            np.testing.assert_allclose(result, desired)

    def test_newline_at_eof_carriage_return(self):
        x = "1 2 3\r\n4 5 6\r\n"
        desired = np.fromstring(x, sep=" ", dtype=np.float32).reshape(-1, 3)
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as f:
            f.write(x)
            f.close()
            reader = CSVReader(f.name)
            result = reader.get_output()[0]
            np.testing.assert_allclose(result, desired)

    def test_holes_in_file(self):
        x = "1 2 3\n4 5 6\n7 8 9"
        desired = np.fromstring(x, sep=" ", dtype=np.float32).reshape(-1, 3)
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as f:
            f.write(x)
            f.close()
            reader = CSVReader(f.name)
            result = reader.get_output()[0]
            np.testing.assert_allclose(result, desired)

    def test_reset(self):
        reader = CSVReader((self.filename1, self.filename2))
        it = reader.iterator()

        data = [chunk for chunk in it]
        it.reset()
        data2 = [chunk for chunk in it]
        np.testing.assert_equal(data, data2)


if __name__ == '__main__':
    unittest.main()



from tqdm._tqdm_notebook import tqdm_notebook


# we just override the default formatting of the widget here
class my_tqdm_notebook(tqdm_notebook):

    @staticmethod
    def status_printer(_, total=None, desc=None, ncols=None):
        # Prepare IPython progress bar
        from ipywidgets import IntProgress, HTML, HBox, Layout, Label

        if total:
            pbar = IntProgress(min=0, max=total)
        else:  # No total? Show info style bar with no progress tqdm status
            pbar = IntProgress(min=0, max=1)
            pbar.value = 1
            pbar.bar_style = 'info'
        if desc:
            description = Label(desc)
            description_box = HBox(children=[description])
            description_box.layout.min_width = '35%'
            description_box.layout.max_width = '35%'
        else:
            description_box = None

        # Prepare status text
        ptext = HTML()
        # Prepare layout
        inner = HBox([pbar, ptext],
                     layout=Layout(padding='0 0 0 20px'))
        # Only way to place text to the right of the bar is to use a container
        box_layout = Layout(display='flex',
                            width='100%')
        container = HBox(children=[description_box, inner] if description_box else [inner],
                         layout=box_layout)

        if ncols is not None:  # use default style of ipywidgets
            # ncols could be 100, "100px", "100%"
            ncols = str(ncols)  # ipywidgets only accepts string
            if ncols[-1].isnumeric():
                # if last value is digit, assume the value is digit
                ncols += 'px'
            pbar.layout.flex = '2'
            container.layout.width = ncols
            container.layout.display = 'inline-flex'
            container.layout.flex_flow = 'row wrap'
        from IPython.core.display import display
        display(container)

        # HTML encoding
        try:  # Py3
            from html import escape
        except ImportError:  # Py2
            from cgi import escape

        def print_status(s='', close=False, bar_style=None, desc=None):
            # Note: contrary to native tqdm, s='' does NOT clear bar
            # goal is to keep all infos if error happens so user knows
            # at which iteration the loop failed.

            # Clear previous output (really necessary?)
            # clear_output(wait=1)

            # Get current iteration value from format_meter string
            if total:
                # n = None
                if s:
                    npos = s.find(r'/|/')  # cause we use bar_format=r'{n}|...'
                    # Check that n can be found in s (else n > total)
                    if npos >= 0:
                        n = int(s[:npos])  # get n from string
                        s = s[npos + 3:]  # remove from string

                        # Update bar with current n value
                        if n is not None:
                            pbar.value = n

            # Print stats
            if s:  # never clear the bar (signal: s='')
                s = s.replace('||', '')  # remove inesthetical pipes
                s = escape(s)  # html escape special characters (like '?')
                ptext.value = s

            # Change bar style
            if bar_style:
                # Hack-ish way to avoid the danger bar_style being overriden by
                # success because the bar gets closed after the error...
                if not (pbar.bar_style == 'danger' and bar_style == 'success'):
                    pbar.bar_style = bar_style

            # Special signal to close the bar
            if close and pbar.bar_style != 'danger':  # hide only if no error
                try:
                    container.close()
                except AttributeError:
                    container.visible = False

            # Update description
            if desc:
                description.value = desc

        return print_status



test_systems = [
                '1FME',
                '2F4K',
                '2JOF',
                '2WAV',
                'A3D',
                'CLN025',
                'GTT',
                'lambda',
                'NuG2',
                'PRB',
                'UVF',
                'NTL9-0-protein_fixed_CC'
]

from glob import glob

import tables
import h5py
import itertools

files = itertools.chain(*[glob('/group/ag_cmb/marscher/feature_sel/*{sys}*{feat}*.h5'.format(sys=s, feat='backbone')) for s in test_systems]
                        )
import tqdm
for f in tqdm.tqdm(files):
    with h5py.File(f) as fh:
        del fh['backbone']



import sys
sys.path.append('/home/mi/marscher/workspace/pyemma/feature_sel/analysis/')
sys.path.append('/home/marscher/feature_sel/analysis/')

from paths import create_traj_top_pairs

import tables
import h5py

import numpy as np
import pyemma
pyemma.config.use_trajectory_lengths_cache = False

test_systems = [
                '1FME',
                '2F4K',
                '2JOF',
                '2WAV',
                'A3D',
                'CLN025',
                'GTT',
                'lambda',
                'NuG2',
                'PRB',
                'UVF',
                'NTL9'
]

#### features
def shrake_ruply(featurizer):
    def featurize(traj):
        import mdtraj
        res = mdtraj.shrake_rupley(traj, probe_radius=0.14, n_sphere_points=960, mode='atom')
        return res

    featurizer.add_custom_func(featurize, dim=featurizer.topology.n_atoms)

def backbone(featurizer):
    featurizer.add_backbone_torsions(cossin=True)

def dist_ca(featurizer):
    featurizer.add_distances_ca()

def res_mindist(featurizer):
    featurizer.add_residue_mindist(scheme='closest-heavy')

def side_sidechain_torsions(featurizer):
    print("adding feature sidechain torssions to featurizer")
    from mdtraj.geometry.dihedral import indices_chi1, indices_chi2, indices_chi3, indices_chi4, indices_chi5, \
        indices_omega
    top = featurizer.topology
    indices = np.vstack((indices_chi1(top),
                         indices_chi2(top),
                         indices_chi3(top),
                         indices_chi4(top),
                         indices_chi5(top),
                         indices_omega(top)))
    assert indices.shape[1] == 4
    from mdtraj import compute_dihedrals

    def compute_side_chains(traj):
        res = compute_dihedrals(traj, indices)
        # cossin
        rad = np.dstack((np.cos(res), np.sin(res)))
        rad = rad.reshape(rad.shape[0], rad.shape[1] * rad.shape[2])
        print('shape chunk:',rad.shape)
        return rad

    featurizer.add_custom_func(compute_side_chains, dim=len(indices)*2)


def dssp(featurizer):
    dim = featurizer.topology.n_residues
    from mdtraj import compute_dssp
    codes = ('H', 'B', 'E', 'G', 'I', 'T', 'S', ' ', 'NA')
    #codes_simple = ('H', 'E', 'C', 'NA')

    mapping = {}
    for i, c in enumerate(codes):
        vec = np.zeros_like(codes, dtype=np.float32)
        vec[i] = 1
        mapping[c] = vec
    #
    #mapping['NA'] = np.zeros_like(codes, dtype=np.float32)

    def _compute(traj):
        result = np.empty((len(traj), traj.topology.n_residues, len(codes)), dtype=np.float32)
        result[:] = np.nan
        assignment = compute_dssp(traj, simplified=False)
        for i, a_ in enumerate(assignment):
            for j, a in enumerate(a_):
                val = mapping.get(a, None)
                if val is None:
                    raise Exception('code at position %s could not be decoded: "%s"' % (i,c))
                result[i, j, :] = val

        result = result.reshape((len(traj), -1))
        assert np.all(np.isfinite(result))
        return result

    featurizer.add_custom_func(_compute, dim=dim*len(codes))


features = (
dssp,
#    side_sidechain_torsions,
#    dist_ca,
#    res_mindist,
#    backbone,
    #shrake_ruply,
)
#### end of features



from sklearn.model_selection import ParameterGrid
import itertools
param_grid = {'test_sys': test_systems,
              'feature': features,
              }

grid = ParameterGrid(param_grid)
# this should be equal to njobs for slurm!
print("num total jobs:", len(list(grid)))

def run(id):
    print('job number: %s' % id)

    try:
        import os
        os.chdir('/home/marscher/NO_BACKUP/data/feature_sel/cached_features/')
        params = grid[id]
        trajs_top = create_traj_top_pairs(test_system=params['test_sys'])
        trajs_flat = list(itertools.chain(*trajs_top['trajs']))
        top = trajs_top['top']
        feature = params['feature']
        feature_name = feature.__name__

        for traj in trajs_flat:
            file_name = "{traj}_{feat}.h5".format(traj=os.path.basename(traj), feat=feature_name)
            print('top: {} \n\ntrajs: {}'.format(top, traj))
            if os.path.exists(file_name):
                print('exists')
                continue
            reader = pyemma.coordinates.source(traj, top=top)
            featurizer = reader.featurizer
            # add feature to featurizer.
            feature(featurizer)

            print("writing to:", file_name)
            with h5py.File(name=file_name, mode='a') as f:
                try:
                    dataset = f.create_dataset(feature_name, shape=(reader.n_frames_total(), reader.ndim),
                                               compression=32001, chunks=True, shuffle=True)
                    dataset[:] = np.nan
                except RuntimeError:
                    print("data set exists already, skipping")
                    return

                t = 0
                with reader.iterator(return_trajindex=False) as it:
                    for chunk in it:
                        dataset[t:t+len(chunk)] = chunk
                        t += len(chunk)

                assert np.all(np.isfinite(dataset[:]))

    except BaseException as e:
        print('bad:', e, id)
        import pdb
        pdb.post_mortem()

    print('successful')


if __name__ == '__main__':
    import os
    #try:os.remove('/group/ag_cmb/marscher/feature_sel/1FME-0-protein-000.dcd_dssp.h5')
    #except:pass
    #run(0)
    import sys
    id = int(sys.argv[1])
    run(id)




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
from contextlib import contextmanager

from unittest.mock import patch
from pyemma.util.files import TemporaryDirectory
from logging import getLogger

import numpy as np
import pyemma.coordinates as coor
import pyemma.util.types as types


logger = getLogger('pyemma.'+'TestCluster')


@contextmanager
def temporary_env(var, value):
    old_val = os.getenv(var, None)
    os.environ[var] = str(value)
    try:
        yield
    finally:
        if old_val is not None:
            os.environ[var] = old_val
        else:
            del os.environ[var]


class TestClusterAssign(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # generate Gaussian mixture
        means = [np.array([-3,0]),
                 np.array([-1,1]),
                 np.array([0,0]),
                 np.array([1,-1]),
                 np.array([4,2])]
        widths = [np.array([0.1,0.1]),
                  np.array([0.1,0.1]),
                  np.array([0.1,0.1]),
                  np.array([0.1,0.1]),
                  np.array([0.1,0.1])]
        # data
        cls.nsample = 1000
        cls.T = len(means)*cls.nsample
        cls.X = np.zeros((cls.T, 2))
        for i in range(len(means)):
            cls.X[i*cls.nsample:(i+1)*cls.nsample,0] = widths[i][0] * np.random.randn() + means[i][0]
            cls.X[i*cls.nsample:(i+1)*cls.nsample,1] = widths[i][1] * np.random.randn() + means[i][1]
        # try assigning actual centers:
        cls.centers = np.array([[-3, 0],
                                [-1, 1],
                                [0, 0],
                                [1, -1],
                                [4, 2]])
        cls.centers_big = cls.X[np.random.choice(cls.nsample, 100)]
        # assignment
        cls.ass = coor.assign_to_centers(data=cls.X, centers=cls.centers, return_dtrajs=False, n_jobs=1)

    def test_chunksize(self):
        assert types.is_int(self.ass.chunksize)

    def test_clustercenters(self):
        c = self.ass
        assert c.clustercenters.shape[0] == self.centers.shape[0]
        assert c.clustercenters.shape[1] == 2

    def test_data_producer(self):
        c = self.ass
        assert c.data_producer is not None

    def test_describe(self):
        c = self.ass
        desc = c.describe()
        assert types.is_string(desc) or types.is_list_of_string(desc)

    def test_dimension(self):
        c = self.ass
        assert types.is_int(c.dimension())
        assert c.dimension() == 1

    def test_dtrajs(self):
        c = self.ass
        assert len(c.dtrajs) == 1
        assert c.dtrajs[0].dtype == c.output_type()
        assert len(c.dtrajs[0]) == self.T
        # assignment in this case should be perfect
        for i in range(self.T):
            assert c.dtrajs[0][i] == int(i / self.nsample)

    def test_return_dtrajs(self):
        dtrajs = coor.assign_to_centers(data=self.X, centers=self.centers)
        for dtraj in dtrajs:
            assert types.is_int_vector(dtraj)

    def test_get_output(self):
        c = self.ass
        O = c.get_output()
        assert types.is_list(O)
        assert len(O) == 1
        assert types.is_int_matrix(O[0])
        assert O[0].shape[0] == self.T
        assert O[0].shape[1] == 1

    def test_in_memory(self):
        c = self.ass
        assert isinstance(c.in_memory, bool)

    def test_iterator(self):
        c = self.ass
        for itraj, chunk in c:
            assert types.is_int(itraj)
            assert types.is_int_matrix(chunk)
            assert chunk.shape[0] <= c.chunksize or c.chunksize == 0
            assert chunk.shape[1] == c.dimension()

    def test_map(self):
        c = self.ass
        Y = c.transform(self.X)
        assert Y.shape[0] == self.T
        assert Y.shape[1] == 1
        # test if consistent with get_output
        assert np.allclose(Y, c.get_output()[0])

    def test_n_frames_total(self):
        c = self.ass
        c.n_frames_total() == self.T

    def test_number_of_trajectories(self):
        c = self.ass
        c.number_of_trajectories() == 1

    def test_output_type(self):
        c = self.ass
        assert c.output_type() == np.int32()

    def test_save_dtrajs(self):
        c = self.ass
        prefix = "test"
        extension = ".dtraj"
        with TemporaryDirectory() as outdir:
            c.save_dtrajs(trajfiles=None, prefix=prefix, output_dir=outdir, extension=extension)

            names = ["%s_%i%s" % (prefix, i, extension)
                     for i in range(c.data_producer.number_of_trajectories())]
            names = [os.path.join(outdir, n) for n in names]

            # check files with given patterns are there
            for f in names:
                os.stat(f)

    def test_trajectory_length(self):
        c = self.ass
        assert c.trajectory_length(0) == self.T
        with self.assertRaises(IndexError):
            c.trajectory_length(1)

    def test_trajectory_lengths(self):
        c = self.ass
        assert len(c.trajectory_lengths()) == 1
        assert c.trajectory_lengths()[0] == c.trajectory_length(0)

    def test_wrong_centers_argument(self):
        dim = 3
        data = np.empty((100,dim))
        centers = np.empty((5, dim+1))

        with self.assertRaises(ValueError):
            c = coor.assign_to_centers(data, centers)

    def test_wrong_centers_argument2(self):
        dim = 3
        data = np.empty((100, dim))
        centers = np.empty(1)

        with self.assertRaises(ValueError):
            c = coor.assign_to_centers(data, centers)

    def test_threads_env_num_threads_fixed(self):
        desired_n_jobs = 2
        with temporary_env('PYEMMA_NJOBS', 0):
            assert os.environ['PYEMMA_NJOBS'] == '0'
            res = coor.assign_to_centers(self.X, self.centers_big, n_jobs=desired_n_jobs, return_dtrajs=False)
            self.assertEqual(res.n_jobs, desired_n_jobs)

    def test_threads_env_num_threads_fixed_def_arg(self):
        """ tests that if no njobs arg is given (None) we fall back to OMP_NUM_THREADS """
        desired_n_jobs = 3
        with temporary_env('PYEMMA_NJOBS', desired_n_jobs):
            assert os.environ['PYEMMA_NJOBS'] == str(desired_n_jobs)
            # note: we want another job number here, but it will be ignored!
            res = coor.assign_to_centers(self.X, self.centers_big, n_jobs=None, return_dtrajs=False)
            self.assertEqual(res.n_jobs, desired_n_jobs)

    def test_threads_omp_env_arg_borked(self):
        """ if the env var can not be interpreted as int, fall back to one thread. """
        expected = 3
        def fake_cpu_count(*args, **kw):
            return expected
        with patch('psutil.cpu_count', fake_cpu_count), temporary_env('PYEMMA_NJOBS', 'this is not right'):
            res = coor.assign_to_centers(self.X, self.centers_big, n_jobs=None, return_dtrajs=False)
            self.assertEqual(res.n_jobs, expected)

    def test_threads_cpu_count_def_arg(self):
        expected = int(os.getenv('PYEMMA_NJOBS', 3))
        def fake_cpu_count(*args, **kw):
            return expected
        with patch('psutil.cpu_count', fake_cpu_count):
            res = coor.assign_to_centers(self.X, self.centers_big, return_dtrajs=False)
        self.assertEqual(res.n_jobs, expected)

    def test_assignment_multithread(self):
        # re-do assignment with multiple threads and compare results
        chunksize = 1000

        assignment_mp = coor.assign_to_centers(self.X, self.centers_big, n_jobs=2, chunksize=chunksize)
        assignment_sp = coor.assign_to_centers(self.X, self.centers_big, n_jobs=1, chunksize=chunksize)

        np.testing.assert_equal(assignment_mp, assignment_sp)

    def test_assignment_multithread_minrsmd(self):
        # re-do assignment with multiple threads and compare results
        import pyemma.datasets as data
        d = data.get_bpti_test_data()
        reader = coor.source(d['trajs'], top=d['top'])

        N_centers = 9
        centers = np.asarray((reader.ra_itraj_jagged[0, [0, 1, 7]],
                              reader.ra_itraj_jagged[1, [32, 1, 23]],
                              reader.ra_itraj_jagged[2, [17, 8, 15]])
                             ).reshape((N_centers, -1))
        chunksize = 1000

        assignment_mp = coor.assign_to_centers(reader, centers, n_jobs=2, chunksize=chunksize, metric='minRMSD')
        assignment_sp = coor.assign_to_centers(reader, centers, n_jobs=1, chunksize=chunksize, metric='minRMSD')

        np.testing.assert_equal(assignment_mp, assignment_sp)

    def test_min_rmsd(self):
        import pyemma.datasets as data
        d = data.get_bpti_test_data()
        reader = coor.source(d['trajs'], top=d['top'])

        N_centers = 9
        centers = np.asarray((reader.ra_itraj_jagged[0, [0, 1, 7]],
                              reader.ra_itraj_jagged[1, [32, 1, 23]],
                              reader.ra_itraj_jagged[2, [17, 8, 15]])
                             ).reshape((N_centers, -1))
        dtraj = coor.assign_to_centers(reader, centers=centers, metric='minRMSD', return_dtrajs=True)

        num_assigned_states = len(np.unique(np.concatenate(dtraj)))
        self.assertEqual(num_assigned_states, N_centers,
                         "assigned states=%s out of %s possible ones."
                         % (num_assigned_states, N_centers))


if __name__ == "__main__":
    unittest.main()



from glob import glob

import tables
import h5py

import os

PRE_CALC_FEATURES_PATH = '/group/ag_cmb/marscher/feature_sel/'

files = glob(PRE_CALC_FEATURES_PATH + '/*.h5')
assert files

to_delete = []
for f in files:
    break
    with h5py.File(f, 'r') as fh:
        if len(fh.keys()) == 0:
            to_delete.append(f)

print(to_delete)

rm = ['/group/ag_cmb/marscher/feature_sel/A3D-0-protein-013.dcd_backbone.h5',
      '/group/ag_cmb/marscher/feature_sel/A3D-1-protein-121.dcd_backbone.h5',
      '/group/ag_cmb/marscher/feature_sel/A3D-1-protein-151.dcd_backbone.h5',
      '/group/ag_cmb/marscher/feature_sel/A3D-1-protein-159.dcd_backbone.h5',
      '/group/ag_cmb/marscher/feature_sel/A3D-0-protein-129.dcd_backbone.h5']

rm = ['/group/ag_cmb/marscher/feature_sel/GTT-0-protein-203.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-079.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-233.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-214.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-066.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-209.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-319.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-143.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-174.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-160.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-031.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-018.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-061.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-191.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-147.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-168.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-052.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-320.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-226.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-325.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-294.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-075.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-308.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-221.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-119.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-229.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-127.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-108.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-015.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-203.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-281.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-059.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-276.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-004.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-029.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-090.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-237.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-040.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-145.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-097.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-118.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-279.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-303.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-114.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-212.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-157.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-190.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-100.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-210.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-035.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-222.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-225.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-221.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-261.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-014.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-201.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-224.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-165.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-158.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-068.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-057.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-101.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-284.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-273.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-191.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-310.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-133.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-251.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-105.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-227.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-207.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-188.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-216.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-234.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-304.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-259.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-169.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-001.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-256.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-048.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-020.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-152.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-264.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-153.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-028.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-302.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-179.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-157.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-197.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-167.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-102.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-212.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-008.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-211.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-300.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-186.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-095.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-175.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-106.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-055.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-116.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-254.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-135.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-287.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-204.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-265.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-033.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-280.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-170.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-188.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-159.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-006.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-163.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-182.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-223.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-271.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-091.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-049.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-241.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-051.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-142.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-318.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-230.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-200.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-196.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-190.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-047.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-231.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-235.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-215.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-136.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-242.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-315.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-138.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-173.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-178.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-292.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-013.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-298.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-130.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-053.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-073.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-081.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-312.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-194.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-208.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-267.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-044.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-240.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-219.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-244.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-208.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-062.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-213.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-184.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-295.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-113.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-286.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-132.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-185.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-148.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-162.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-104.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-323.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-002.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-313.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-000.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-178.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-046.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-111.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-151.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-285.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-206.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-195.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-199.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-093.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-250.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-255.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-180.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-110.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-243.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-234.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-214.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-249.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-163.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-094.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-024.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-213.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-071.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-238.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-165.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-050.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-043.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-317.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-065.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-058.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-023.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-171.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-253.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-232.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-034.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-322.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-176.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-139.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-083.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-270.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-140.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-228.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-210.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-187.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-283.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-222.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-172.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-084.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-215.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-324.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-164.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-177.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-220.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-277.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-299.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-181.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-016.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-189.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-005.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-236.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-205.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-122.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-082.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-241.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-154.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-193.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-168.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-092.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-150.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-248.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-209.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-078.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-089.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-278.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-167.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-022.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-103.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-125.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-038.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-027.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-211.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-039.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-217.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-227.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-232.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-195.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-258.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-010.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-174.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-171.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-060.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-305.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-262.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-137.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-291.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-238.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-141.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-076.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-054.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-126.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-164.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-072.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-129.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-069.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-099.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-199.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-032.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-245.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-282.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-183.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-201.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-107.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-219.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-196.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-252.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-189.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-041.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-217.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-289.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-169.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-128.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-056.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-202.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-070.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-030.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-207.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-045.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-204.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-229.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-067.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-124.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-301.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-236.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-161.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-275.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-170.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-269.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-063.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-007.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-231.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-146.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-309.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-235.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-218.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-155.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-240.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-237.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-263.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-268.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-086.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-186.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-257.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-272.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-064.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-175.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-003.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-297.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-019.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-243.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-077.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-120.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-026.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-172.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-194.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-042.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-112.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-246.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-085.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-228.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-307.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-197.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-180.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-198.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-288.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-266.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-225.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-173.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-296.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-290.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-192.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-021.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-098.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-316.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-087.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-242.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-144.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-224.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-088.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-260.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-011.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-230.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-183.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-200.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-185.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-017.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-121.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-131.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-036.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-149.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-233.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-096.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-012.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-080.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-166.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-176.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-198.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-223.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-274.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-226.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-025.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-321.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-134.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-009.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-306.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-247.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-117.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-109.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-166.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-115.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-293.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-220.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-177.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-205.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-187.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-239.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-314.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-193.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-123.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-206.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-202.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-311.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-216.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-037.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-074.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-239.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-192.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-0-protein-184.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-218.dcd_dssp.h5',
      '/group/ag_cmb/marscher/feature_sel/GTT-1-protein-156.dcd_dssp.h5']

for f in rm:
    os.remove(f)



# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
Created on 30.04.2015

@author: marscher
'''


from tempfile import NamedTemporaryFile

import os
import tempfile
import unittest

from unittest import mock

from pyemma.coordinates import api
from pyemma.coordinates.data.feature_reader import FeatureReader
from pyemma.coordinates.data.numpy_filereader import NumPyFileReader
from pyemma.coordinates.data.py_csv_reader import PyCSVReader
from pyemma.coordinates.data.util.traj_info_backends import SqliteDB
from pyemma.coordinates.data.util.traj_info_cache import TrajectoryInfoCache
from pyemma.coordinates.tests.util import create_traj
from pyemma.datasets import get_bpti_test_data
from pyemma.util import config
from pyemma.util.contexts import settings
from pyemma.util.files import TemporaryDirectory
import mdtraj
import pkg_resources
import pyemma
import numpy as np

xtcfiles = get_bpti_test_data()['trajs']
pdbfile = get_bpti_test_data()['top']


class TestTrajectoryInfoCache(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.old_instance = TrajectoryInfoCache.instance()
        config.use_trajectory_lengths_cache = True

    def setUp(self):
        self.work_dir = tempfile.mkdtemp(prefix="traj_cache_test")
        self.tmpfile = tempfile.mktemp(dir=self.work_dir)
        self.db = TrajectoryInfoCache(self.tmpfile)

        # overwrite TrajectoryInfoCache._instance with self.db...
        TrajectoryInfoCache._instance = self.db
        config.use_trajectory_lengths_cache = True

    def tearDown(self):
        self.db.close()
        os.unlink(self.tmpfile)

        import shutil
        shutil.rmtree(self.work_dir, ignore_errors=True)

    @classmethod
    def tearDownClass(cls):
        TrajectoryInfoCache._instance = cls.old_instance
        config.use_trajectory_lengths_cache = False

    def test_get_instance(self):
        # test for exceptions in singleton creation
        inst = TrajectoryInfoCache.instance()
        inst.current_db_version
        self.assertIs(inst, self.db)

    def test_store_load_traj_info(self):
        x = np.random.random((10, 3))
        from pyemma.util._config import Config
        my_conf = Config()
        my_conf.cfg_dir = self.work_dir
        with mock.patch('pyemma.coordinates.data.util.traj_info_cache.config', my_conf):
            with NamedTemporaryFile(delete=False) as fh:
                np.savetxt(fh.name, x)
                reader = api.source(fh.name)
                info = self.db[fh.name, reader]
                self.db.close()
                self.db.__init__(self.db._database.filename)
                info2 = self.db[fh.name, reader]
                self.assertEqual(info2, info)

    def test_exceptions(self):
        # in accessible files
        not_existant = ''.join(
            chr(i) for i in np.random.randint(65, 90, size=10)) + '.npy'
        bad = [not_existant]  # should be unaccessible or non existent
        with self.assertRaises(ValueError) as cm:
            api.source(bad)
            assert bad[0] in cm.exception.message

        # empty files
        with NamedTemporaryFile(delete=False) as f:
            f.close()
            with self.assertRaises(ValueError) as cm:
                api.source(f.name)
                assert f.name in cm.exception.message

        # bogus files
        with NamedTemporaryFile(suffix='.npy', delete=False) as f:
            x = np.array([1, 2, 3])
            np.save(f, x)
            with open(f.name, 'wb') as f2:
                f2.write(b'asdf')
            with self.assertRaises(IOError) as cm:
                api.source(f.name)

    def test_featurereader_xtc(self):
        # cause cache failures
        with settings(use_trajectory_lengths_cache=False):
            reader = FeatureReader(xtcfiles, pdbfile)

        results = {}
        for f in xtcfiles:
            traj_info = self.db[f, reader]
            results[f] = traj_info.ndim, traj_info.length, traj_info.offsets

        expected = {}
        for f in xtcfiles:
            with mdtraj.open(f) as fh:
                length = len(fh)
                ndim = fh.read(1)[0].shape[1]
                offsets = fh.offsets if hasattr(fh, 'offsets') else []
                expected[f] = ndim, length, offsets

        np.testing.assert_equal(results, expected)

    def test_npy_reader(self):
        lengths_and_dims = [(7, 3), (23, 3), (27, 3)]
        data = [
            np.empty((n, dim)) for n, dim in lengths_and_dims]
        files = []
        with TemporaryDirectory() as td:
            for i, x in enumerate(data):
                fn = os.path.join(td, "%i.npy" % i)
                np.save(fn, x)
                files.append(fn)

            reader = NumPyFileReader(files)

            # cache it and compare
            results = {f: (self.db[f, reader].length, self.db[f, reader].ndim,
                           self.db[f, reader].offsets) for f in files}
            expected = {f: (len(data[i]), data[i].shape[1], [])
                        for i, f in enumerate(files)}
            np.testing.assert_equal(results, expected)

    def test_csvreader(self):
        data = np.random.random((101, 3))
        fn = tempfile.mktemp()
        try:
            np.savetxt(fn, data)
            # calc offsets
            offsets = [0]
            with open(fn, PyCSVReader.DEFAULT_OPEN_MODE) as new_fh:
                while new_fh.readline():
                    offsets.append(new_fh.tell())
            reader = PyCSVReader(fn)
            assert reader.dimension() == 3
            trajinfo = reader._get_traj_info(fn)
            np.testing.assert_equal(offsets, trajinfo.offsets)
        finally:
            os.unlink(fn)

    def test_fragmented_reader(self):
        top_file = pkg_resources.resource_filename(__name__, 'data/test.pdb')
        trajfiles = []
        nframes = []
        with TemporaryDirectory() as wd:
            for _ in range(3):
                f, _, l = create_traj(top_file, dir=wd)
                trajfiles.append(f)
                nframes.append(l)
            # three trajectories: one consisting of all three, one consisting of the first,
            # one consisting of the first and the last
            reader = api.source(
                [trajfiles, [trajfiles[0]], [trajfiles[0], trajfiles[2]]], top=top_file)
            np.testing.assert_equal(reader.trajectory_lengths(),
                                    [sum(nframes), nframes[0], nframes[0] + nframes[2]])

    def test_feature_reader_xyz(self):
        traj = mdtraj.load(xtcfiles, top=pdbfile)
        length = len(traj)

        with NamedTemporaryFile(mode='wb', suffix='.xyz', delete=False) as f:
            fn = f.name
            traj.save_xyz(fn)
            f.close()
            reader = pyemma.coordinates.source(fn, top=pdbfile)
            self.assertEqual(reader.trajectory_length(0), length)

    def test_data_in_mem(self):
        # make sure cache is not used for data in memory!
        data = [np.empty((3, 3))] * 3
        api.source(data)
        self.assertEqual(self.db.num_entries, 0)

    def test_old_db_conversion(self):
        # prior 2.1, database only contained lengths (int as string) entries
        # check conversion is happening
        with NamedTemporaryFile(suffix='.npy', delete=False) as f:
            db = TrajectoryInfoCache(None)
            fn = f.name
            np.save(fn, [1, 2, 3])
            f.close()  # windows sucks
            reader = api.source(fn)
            hash = db._get_file_hash(fn)
            from pyemma.coordinates.data.util.traj_info_backends import DictDB
            db._database = DictDB()
            db._database.db_version = 0

            info = db[fn, reader]
            assert info.length == 3
            assert info.ndim == 1
            assert info.offsets == []

    def test_corrupted_db(self):
        with NamedTemporaryFile(mode='w', suffix='.dat', delete=False) as f:
            f.write("makes no sense!!!!")
            f.close()
        name = f.name
        import warnings
        with warnings.catch_warnings(record=True) as cm:
            warnings.simplefilter('always')
            db = TrajectoryInfoCache(name)
            assert len(cm) == 1
            assert "corrupted" in str(cm[-1].message)

        # ensure we can perform lookups on the broken db without exception.
        r = api.source(xtcfiles[0], top=pdbfile)
        db[xtcfiles[0], r]

    def test_n_entries(self):
        assert config.use_trajectory_lengths_cache
        self.assertEqual(self.db.num_entries, 0)
        assert TrajectoryInfoCache._instance is self.db
        pyemma.coordinates.source(xtcfiles, top=pdbfile)
        self.assertEqual(self.db.num_entries, len(xtcfiles))

    def test_max_n_entries(self):
        data = [np.random.random((10, 3)) for _ in range(20)]
        max_entries = 10
        config.traj_info_max_entries = max_entries
        files = []
        with TemporaryDirectory() as td:
            for i, arr in enumerate(data):
                f = os.path.join(td, "%s.npy" % i)
                np.save(f, arr)
                files.append(f)
            pyemma.coordinates.source(files)
        self.assertLessEqual(self.db.num_entries, max_entries)
        self.assertGreater(self.db.num_entries, 0)

    def test_max_size(self):
        data = [np.random.random((150, 10)) for _ in range(150)]
        max_size = 1

        files = []
        with TemporaryDirectory() as td, settings(traj_info_max_size=max_size, show_progress_bars=False):
            for i, arr in enumerate(data):
                f = os.path.join(td, "%s.txt" % i)
                # save as txt to enforce creation of offsets
                np.savetxt(f, arr)
                files.append(f)
            pyemma.coordinates.source(files)

        self.assertLessEqual(os.stat(self.db.database_filename).st_size / 1024, config.traj_info_max_size)
        self.assertGreater(self.db.num_entries, 0)

    def test_no_working_directory(self):
        # this is the case as long as the user has not yet created a config directory via config.save()
        self.db._database = SqliteDB(filename=None)

        # trigger caching
        pyemma.coordinates.source(xtcfiles, top=pdbfile)

    def test_no_sqlite(self):
        # create new instance (init has to be called, install temporary import hook to raise importerror for sqlite3
        import sys
        del sys.modules['sqlite3']

        class meta_ldr(object):
            def find_module(self, fullname, path):
                if fullname.startswith('sqlite3'):
                    return self

            def load_module(self, fullname, path=None):
                raise ImportError()

        import warnings
        try:
            sys.meta_path.insert(0, meta_ldr())
            # import sqlite3
            with warnings.catch_warnings(record=True) as cw:
                db = TrajectoryInfoCache()
                self.assertNotIsInstance(db._database, SqliteDB)
            self.assertEqual(len(cw), 1)
            self.assertIn("sqlite3 package not available", cw[0].message.args[0])
        finally:
            del sys.meta_path[0]

    def test_in_memory_db(self):
        """ new instance, not yet saved to disk, no lru cache avail """
        old_cfg_dir = config.cfg_dir
        try:
            config._cfg_dir = ''
            db = TrajectoryInfoCache()
            reader = pyemma.coordinates.source(xtcfiles, top=pdbfile)

            info = db[xtcfiles[0], reader]
            self.assertIsInstance(db._database, SqliteDB)

            directory = db._database._database_from_key(info.hash_value)
            assert directory is None
        finally:
            from pyemma.util.exceptions import ConfigDirectoryException
            try:
                config.cfg_dir = old_cfg_dir
            except ConfigDirectoryException:
                pass

    def test_stress(self):
        arrays = [np.empty((5, 2))] * 100
        npy_files = [os.path.join(self.work_dir, '{}.npy'.format(i)) for i in range(len(arrays))]
        [np.save(f, x) for f, x in zip(npy_files, arrays)]
        env = os.environ.copy()
        env['PYEMMA_CFG_DIR'] = self.work_dir
        import subprocess
        import sys
        import time
        script = 'import pyemma; pyemma.coordinates.source({files})' \
            .format(cfg_dir=self.work_dir, files=npy_files)
        failed = False
        procs = [subprocess.Popen([sys.executable, '-c', script], env=env) for _ in range(10)]
        error = None
        while procs:
            for proc in procs:
                retcode = proc.poll()
                if retcode is not None:
                    if retcode != 0:
                        pass
                        #stdout = proc.stdout.read()
                        #stderr = proc.stderr.read()
                        #error = '{};;{}'.format(stdout, stderr)
                    procs.remove(proc)
                    #break
                else:  # No process is done, wait a bit and check again.
                    time.sleep(.1)
                    continue

            # Here, `proc` has finished with return code `retcode`
            if retcode is not None and retcode != 0:
                print('process failed with {}'.format(retcode))
                failed = True
                break

        self.assertTrue(not failed, msg=error)

if __name__ == "__main__":
    unittest.main()




def get_logger(name, filename):
    import logging, sys
    logger = logging.getLogger('cov_')
    logger.addHandler(logging.FileHandler(filename))
    logger.addHandler(logging.StreamHandler(stream=sys.stdout))
    logger.setLevel('INFO')
    for h in logger.handlers:
        h.add




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import numpy as _np
from pyemma.util.types import  is_iterable_of_int as _is_iterable_of_int, is_int as _is_int

__author__ = 'noe'


def plot_implied_timescales(ITS, ax=None, outfile=None, show_mle=True, show_mean=True,
                            xlog=False, ylog=True, confidence=0.95, refs=None, nits=-1,
                            process=None, units='steps', dt=1., **kwargs):
    r"""Implied timescale plot

    Parameters
    ----------
    ITS : implied timescales object.
        Object whose data will be plotted. Must provide the functions: get_timescales() and get_timescales(i) where i is the
        the property samples_available
    ax : matplotlib Axes object, optional, default = None
        the axes to plot to. When set to None the default Axes object will be used.
    outfile : str, optional, default = None
        output file to write the figure to. When not given, the plot will be displayed
    show_mean : bool, default = True
        Line for mean value will be shown, if available
    show_mle : bool, default = True
        Line for maximum likelihood estimate will be shown
    xlog : bool, optional, default = False
        Iff true, the x-Axis is logarithmized
    ylog : bool, optional, default = True
        Iff true, the y-Axis is logarithmized
    confidence : float, optional, default = 0.95
        The confidence interval for plotting error bars (if available)
    refs : ndarray((m), dtype=float), optional, default = None
        Reference (exact solution or other reference) timescales if known. The number of timescales must match those
        in the ITS object
    nits: integer, default = -1
        Number of implied timescales to be shown. The default behaviour (-1) is to show all timescales available.
        :py:obj:`nits` != -1 and :py:obj:`process` != None are mutually exclusive
    process : iterable of integers, default is None
        list or ndarray((m), dtype=int) containing a list of the processes to be shown. The default behaviour is
        to show all timescales available.
        :py:obj:`process` != None and :py:obj:`nits` != -1 are mutually exclusive
    units: str or list (len=2) of strings, optional, default = 'steps'
        Affects the labeling of the axes. Used with :py:obj:`dt`, allows for changing the physical units of the axes.
        Accepts simple LaTeX math strings, eg. '$\mu$s'
        If this parameter is a list, it will be assumed that units[0] is for the x-axis and units[1] for the y-axis.
    dt: float or list(len=2) of floats, optional, default = 1.0
        Physical time between frames, expressed the units given in :py:obj:`units`. E.g, if you know that each
        frame corresponds to .010 ns, you can use the combination of parameters :py:obj:`dt` =0.01,
        :py:obj:`units` ='ns' to display the implied timescales in ns (instead of frames)
        If this parameter is a list, it will be assumed that dt[0] is for the x-axis and dt[1] for the y-axis.

    **kwargs: Will be passed to pyplot.plot when plotting the MLE datapoints (not the bootstrapped means).
            See the doc of pyplot for more options. Most useful lineproperties like `marker='o'` and/or :markersize=5

    Returns
    -------
    ax : Axes object containing the plot

    """
    import matplotlib.pyplot as _plt
    # check input
    if ax is None:
        ax = _plt.gca()
    colors = ['blue', 'red', 'green', 'cyan', 'purple', 'orange', 'violet']
    lags = ITS.lagtimes
    xmax = _np.max(lags)
    srt = _np.argsort(lags)
    # Check the processes to be shown
    if process is not None:
        if nits != -1:
            raise TypeError(
                'optional arguments nits and process are mutually exclusive:', nits, process)
        if not _is_iterable_of_int(process):
            raise ValueError('process has to be an iterable of integers')
        if _np.max(process)+1 > ITS.number_of_timescales:
            raise ValueError(
                'requested process %u, whereas ITS only contains %u timescales' % (
                    _np.max(process), ITS.number_of_timescales))
        # Now that it's for sure that nits==-1, process is iter_of_ints, and the requested processes exist in its object:
        its_idx = process
    else:
        if not _is_int(nits):
            raise TypeError('nits is not an integer, ',nits)
        if nits == -1:
            nits = ITS.number_of_timescales
        its_idx = _np.arange(ITS.number_of_timescales)[:nits]
    # Check units and dt for user error.
    if isinstance(units, list) and len(units) != 2:
        raise TypeError("If units is a list, len(units) has to be = 2")
    if isinstance(dt, list) and len(dt) != 2:
        raise TypeError("If dt is a list, len(dt) has to be = 2")
    # Create list of units and dts for different axis
    if isinstance(units, str):
        units = [units] * 2
    if isinstance(dt, (float, int)):
        dt = [dt] * 2
    #ymin = min(_np.min(lags), _np.min(ITS.get_timescales()))
    #ymax = 1.5*_np.min(ITS.get_timescales())
    for i in its_idx:
        # plot estimate
        if show_mle:
            ax.plot(
                lags[srt] * dt[0], ITS.get_timescales(process=i)[srt] * dt[1],
                color=colors[i % len(colors)], **kwargs)
        # sample available?
        if ITS.samples_available:# and ITS.sample_number_of_timescales > i):
            # plot sample mean
            if show_mean:
                ax.plot(
                    lags[srt] * dt[0], ITS.get_sample_mean(process=i)[srt] * dt[1], marker='o',
                    color=colors[i % len(colors)], linestyle='dashed')
            (lconf, rconf) = ITS.get_sample_conf(confidence, i)
            ax.fill_between(
                lags[srt] * dt[0], lconf[srt] * dt[1], rconf[srt] * dt[1],
                alpha=0.2, color=colors[i % len(colors)])
        # reference available?
        if refs is not None:
            tref = refs[i] * dt[1]
            ax.plot([0, min(tref, xmax) * dt[0]], [tref, tref], color='black', linewidth=1)
    # cutoff
    ax.plot(lags[srt] * dt[0], lags[srt] * dt[1], linewidth=2, color='black')
    ax.set_xlim([1.0 * dt[0], xmax * dt[0]])
    #ax.set_ylim([ymin,ymax])
    ax.fill_between(
        lags[srt] * dt[0], ax.get_ylim()[0]*_np.ones(len(lags))*dt[1], lags[srt] * dt[1],
        alpha=0.5, color='grey')
    # formatting
    ax.set_xlabel('lag time / %s' % units[0])
    ax.set_ylabel('timescale / %s' % units[1])
    if (xlog):
        ax.set_xscale('log')
    if (ylog):
        ax.set_yscale('log')
    # show or save
    # if outfile is None:
    #    _plt.show()
    if outfile is not None:
        _plt.savefig(outfile)
    return ax



import numpy as _np
from ..util import ZeroRankError as _ZeroRankError

__author__ = 'noe'


def sort_by_norm(evals, evecs):
    """
    Sorts the eigenvalues and eigenvectors by descending norm of the eigenvalues

    Parameters
    ----------
    evals: ndarray(n)
        eigenvalues
    evecs: ndarray(n,n)
        eigenvectors in a column matrix

    Returns
    -------
    (evals, evecs) : ndarray(m), ndarray(n,m)
        the sorted eigenvalues and eigenvectors

    """
    # norms
    evnorms = _np.abs(evals)
    # sort
    I = _np.argsort(evnorms)[::-1]
    # permute
    evals2 = evals[I]
    evecs2 = evecs[:, I]
    # done
    return evals2, evecs2


def spd_eig(W, epsilon=1e-10, method='QR', canonical_signs=False):
    """ Rank-reduced eigenvalue decomposition of symmetric positive definite matrix.

    Removes all negligible eigenvalues

    Parameters
    ----------
    W : ndarray((n, n), dtype=float)
        Symmetric positive-definite (spd) matrix.
    epsilon : float
        Truncation parameter. Eigenvalues with norms smaller than this cutoff will
        be removed.
    method : str
        Method to perform the decomposition of :math:`W` before inverting. Options are:

        * 'QR': QR-based robust eigenvalue decomposition of W
        * 'schur': Schur decomposition of W
    canonical_signs : boolean, default = False
        Fix signs in V, s. t. the largest element of in every row of V is positive.

    Returns
    -------
    s : ndarray(k)
        k non-negligible eigenvalues, sorted by descending norms

    V : ndarray(n, k)
        k leading eigenvectors
    """
    # check input
    assert _np.allclose(W.T, W), 'W is not a symmetric matrix'

    if method.lower() == 'qr':
        from .eig_qr.eig_qr import eig_qr
        s, V = eig_qr(W)
    # compute the Eigenvalues of C0 using Schur factorization
    elif method.lower() == 'schur':
        from scipy.linalg import schur
        S, V = schur(W)
        s = _np.diag(S)
    else:
        raise ValueError('method not implemented: ' + method)

    s, V = sort_by_norm(s, V) # sort them

    # determine the cutoff. We know that C0 is an spd matrix,
    # so we select the truncation threshold such that everything that is negative vanishes
    evmin = _np.min(s)
    if evmin < 0:
        epsilon = max(epsilon, -evmin + 1e-16)

    # determine effective rank m and perform low-rank approximations.
    evnorms = _np.abs(s)
    n = _np.shape(evnorms)[0]
    m = n - _np.searchsorted(evnorms[::-1], epsilon)
    if m == 0:
        raise _ZeroRankError('All eigenvalues are smaller than %g, rank reduction would discard all dimensions.'%epsilon)
    Vm = V[:, 0:m]
    sm = s[0:m]

    if canonical_signs:
        # enforce canonical eigenvector signs
        for j in range(m):
            jj = _np.argmax(_np.abs(Vm[:, j]))
            Vm[:, j] *= _np.sign(Vm[jj, j])

    return sm, Vm


def spd_inv(W, epsilon=1e-10, method='QR'):
    """
    Compute matrix inverse of symmetric positive-definite matrix :math:`W`.

    by first reducing W to a low-rank approximation that is truly spd
    (Moore-Penrose inverse).

    Parameters
    ----------
    W : ndarray((m,m), dtype=float)
        Symmetric positive-definite (spd) matrix.
    epsilon : float
        Truncation parameter. Eigenvalues with norms smaller than this cutoff will
        be removed.
    method : str
        Method to perform the decomposition of :math:`W` before inverting. Options are:

        * 'QR': QR-based robust eigenvalue decomposition of W
        * 'schur': Schur decomposition of W

    Returns
    -------
    L : ndarray((n, r))
        the Moore-Penrose inverse of the symmetric positive-definite matrix :math:`W`

    """
    if (_np.shape(W)[0] == 1):
        if W[0,0] < epsilon:
            raise _ZeroRankError(
                'All eigenvalues are smaller than %g, rank reduction would discard all dimensions.' % epsilon)
        Winv = 1./W[0,0]
    else:
        sm, Vm = spd_eig(W, epsilon=epsilon, method=method)
        Winv = _np.dot(Vm, _np.diag(1.0 / sm)).dot(Vm.T)

    # return split
    return Winv


def spd_inv_sqrt(W, epsilon=1e-10, method='QR', return_rank=False):
    """
    Computes :math:`W^{-1/2}` of symmetric positive-definite matrix :math:`W`.

    by first reducing W to a low-rank approximation that is truly spd.

    Parameters
    ----------
    W : ndarray((m,m), dtype=float)
        Symmetric positive-definite (spd) matrix.
    epsilon : float
        Truncation parameter. Eigenvalues with norms smaller than this cutoff will
        be removed.
    method : str
        Method to perform the decomposition of :math:`W` before inverting. Options are:

        * 'QR': QR-based robust eigenvalue decomposition of W
        * 'schur': Schur decomposition of W

    Returns
    -------
    L : ndarray((n, r))
        :math:`W^{-1/2}` after reduction of W to a low-rank spd matrix

    """
    if _np.shape(W)[0] == 1:
        if W[0,0] < epsilon:
            raise _ZeroRankError(
                'All eigenvalues are smaller than %g, rank reduction would discard all dimensions.' % epsilon)
        Winv = 1./_np.sqrt(W[0, 0])
        sm = _np.ones(1)
    else:
        sm, Vm = spd_eig(W, epsilon=epsilon, method=method)
        Winv = _np.dot(Vm, _np.diag(1.0 / _np.sqrt(sm))).dot(Vm.T)

    # return split
    if return_rank:
        return Winv, sm.shape[0]
    else:
        return Winv


def spd_inv_split(W, epsilon=1e-10, method='QR', canonical_signs=False):
    """
    Compute :math:`W^{-1} = L L^T` of the symmetric positive-definite matrix :math:`W`.

    by first reducing W to a low-rank approximation that is truly spd.

    Parameters
    ----------
    W : ndarray((m,m), dtype=float)
        Symmetric positive-definite (spd) matrix.
    epsilon : float
        Truncation parameter. Eigenvalues with norms smaller than this cutoff will
        be removed.
    method : str
        Method to perform the decomposition of :math:`W` before inverting. Options are:

        * 'QR': QR-based robust eigenvalue decomposition of W
        * 'schur': Schur decomposition of W

     canonical_signs : boolean, default = False
        Fix signs in L, s. t. the largest element of in every row of L is positive.

    Returns
    -------
    L : ndarray((n, r))
        Matrix :math:`L` from the decomposition :math:`W^{-1} = L L^T`.

    """
    if (_np.shape(W)[0] == 1):
        if W[0,0] < epsilon:
            raise _ZeroRankError(
                'All eigenvalues are smaller than %g, rank reduction would discard all dimensions.' % epsilon)
        L = 1./_np.sqrt(W[0,0])
    else:
        sm, Vm = spd_eig(W, epsilon=epsilon, method=method, canonical_signs=canonical_signs)
        L = _np.dot(Vm, _np.diag(1.0/_np.sqrt(sm)))

    # return split
    return L


def eig_corr(C0, Ct, epsilon=1e-10, method='QR', sign_maxelement=False):
    r""" Solve generalized eigenvalue problem with correlation matrices C0 and Ct

    Numerically robust solution of a generalized Hermitian (symmetric) eigenvalue
    problem of the form

    .. math::
        \mathbf{C}_t \mathbf{r}_i = \mathbf{C}_0 \mathbf{r}_i l_i

    Computes :math:`m` dominant eigenvalues :math:`l_i` and eigenvectors
    :math:`\mathbf{r}_i`, where :math:`m` is the numerical rank of the problem.
    This is done by first conducting a Schur decomposition of the symmetric
    positive matrix :math:`\mathbf{C}_0`, then truncating its spectrum to
    retain only eigenvalues that are numerically greater than zero, then using
    this decomposition to define an ordinary eigenvalue Problem for
    :math:`\mathbf{C}_t` of size :math:`m`, and then solving this eigenvalue
    problem.

    Parameters
    ----------
    C0 : ndarray (n,n)
        time-instantaneous correlation matrix. Must be symmetric positive definite
    Ct : ndarray (n,n)
        time-lagged correlation matrix. Must be symmetric
    epsilon : float
        eigenvalue norm cutoff. Eigenvalues of C0 with norms <= epsilon will be
        cut off. The remaining number of Eigenvalues define the size of
        the output.
    method : str
        Method to perform the decomposition of :math:`W` before inverting. Options are:

        * 'QR': QR-based robust eigenvalue decomposition of W
        * 'schur': Schur decomposition of W
    sign_maxelement : bool
        If True, re-scale each eigenvector such that its entry with maximal absolute value
        is positive.


    Returns
    -------
    l : ndarray (m)
        The first m generalized eigenvalues, sorted by descending norm
    R : ndarray (n,m)
        The first m generalized eigenvectors, as a column matrix.

    """
    L = spd_inv_split(C0, epsilon=epsilon, method=method, canonical_signs=True)
    Ct_trans = _np.dot(_np.dot(L.T, Ct), L)

    # solve the symmetric eigenvalue problem in the new basis
    if _np.allclose(Ct.T, Ct):
        from scipy.linalg import eigh
        l, R_trans = eigh(Ct_trans)
    else:
        from scipy.linalg import eig
        l, R_trans = eig(Ct_trans)

    # sort eigenpairs
    l, R_trans = sort_by_norm(l, R_trans)

    # transform the eigenvectors back to the old basis
    R = _np.dot(L, R_trans)

    # Change signs of eigenvectors:
    if sign_maxelement:
        for j in range(R.shape[1]):
            imax = _np.argmax(_np.abs(R[:, j]))
            R[:, j] *= _np.sign(R[imax, j])

    # return result
    return l, R




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""This module implements IO and manipulation function for discrete trajectories

Discrete trajectories are generally ndarrays of type integer
We store them either as single column ascii files or as ndarrays of shape (n,) in binary .npy format.

.. moduleauthor:: B. Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>
.. moduleauthor:: F. Noe <frank DOT noe AT fu-berlin DOT de>

"""
import numpy as np
from pyemma.util.types import ensure_dtraj_list as _ensure_dtraj_list
from pyemma.util.annotators import shortcut


__author__ = 'noe'

################################################################################
# ascii
################################################################################

@shortcut('read_dtraj')
def read_discrete_trajectory(filename):
    """Read discrete trajectory from ascii file.

    The ascii file containing a single column with integer entries is
    read into an array of integers.

    Parameters
    ----------
    filename : str
        The filename of the discrete state trajectory file.
        The filename can either contain the full or the
        relative path to the file.

    Returns
    -------
    dtraj : (M, ) ndarray
        Discrete state trajectory.

    """
    with open(filename, "r") as f:
        lines=f.read()
        dtraj=np.fromstring(lines, dtype=int, sep="\n")
        return dtraj


@shortcut('write_dtraj')
def write_discrete_trajectory(filename, dtraj):
    r"""Write discrete trajectory to ascii file.

    The discrete trajectory is written to a
    single column ascii file with integer entries

    Parameters
    ----------
    filename : str
        The filename of the discrete state trajectory file.
        The filename can either contain the full or the
        relative path to the file.

    dtraj : array-like
        Discrete state trajectory.

    """
    dtraj=np.asarray(dtraj)
    with open(filename, 'w') as f:
        dtraj.tofile(f, sep='\n', format='%d')

################################################################################
# binary
################################################################################

@shortcut('load_dtraj')
def load_discrete_trajectory(filename):
    r"""Read discrete trajectory form binary file.

    The binary file is a one dimensional numpy array
    of integers stored in numpy .npy format.

    Parameters
    ----------
    filename : str
        The filename of the discrete state trajectory file.
        The filename can either contain the full or the
        relative path to the file.

    Returns
    -------
    dtraj : (M,) ndarray
        Discrete state trajectory.

    """
    dtraj=np.load(filename)
    return dtraj


@shortcut('save_dtraj')
def save_discrete_trajectory(filename, dtraj):
    r"""Write discrete trajectory to binary file.

    The discrete trajectory is stored as ndarray of integers
    in numpy .npy format.

    Parameters
    ----------
    filename : str
        The filename of the discrete state trajectory file.
        The filename can either contain the full or the
        relative path to the file.


    dtraj : array-like
        Discrete state trajectory.

    """
    dtraj=np.asarray(dtraj)
    np.save(filename, dtraj)


################################################################################
# simple statistics
################################################################################


@shortcut('histogram')
def count_states(dtrajs, ignore_negative=False):
    r"""returns a histogram count

    Parameters
    ----------
    dtrajs : array_like or list of array_like
        Discretized trajectory or list of discretized trajectories
    ignore_negative, bool, default=False
        Ignore negative elements. By default, a negative element will cause an
        exception

    Returns
    -------
    count : ndarray((n), dtype=int)
        the number of occurrences of each state. n=max+1 where max is the largest state index found.

    """
    # format input
    dtrajs = _ensure_dtraj_list(dtrajs)
    # make bincounts for each input trajectory
    nmax = 0
    bcs = []
    for dtraj in dtrajs:
        if ignore_negative:
            dtraj = dtraj[np.where(dtraj >= 0)]
        bc = np.bincount(dtraj)
        nmax = max(nmax, bc.shape[0])
        bcs.append(bc)
    # construct total bincount
    res = np.zeros(nmax, dtype=int)
    # add up individual bincounts
    for i, bc in enumerate(bcs):
        res[:bc.shape[0]] += bc
    return res


def visited_set(dtrajs):
    r"""returns the set of states that have at least one count

    Parameters
    ----------
    dtraj : array_like or list of array_like
        Discretized trajectory or list of discretized trajectories

    Returns
    -------
    vis : ndarray((n), dtype=int)
        the set of states that have at least one count.
    """
    hist = count_states(dtrajs)
    return np.argwhere(hist > 0)[:,0]


@shortcut('nstates')
def number_of_states(dtrajs, only_used = False):
    r"""returns the number of states in the given trajectories.

    Parameters
    ----------
    dtraj : array_like or list of array_like
        Discretized trajectory or list of discretized trajectories
    only_used = False : boolean
        If False, will return max+1, where max is the largest index used.
        If True, will return the number of states that occur at least once.
    """
    dtrajs = _ensure_dtraj_list(dtrajs)
    if only_used:
        # only states with counts > 0 wanted. Make a bincount and count nonzeros
        bc = count_states(dtrajs)
        return np.count_nonzero(bc)
    else:
        # all states wanted, included nonpopulated ones. return max + 1
        imax = 0
        for dtraj in dtrajs:
            imax = max(imax, np.max(dtraj))
        return imax+1

################################################################################
# indexing
################################################################################


def index_states(dtrajs, subset=None):
    """Generates a trajectory/time indexes for the given list of states

    Parameters
    ----------
    dtraj : array_like or list of array_like
        Discretized trajectory or list of discretized trajectories. Negative elements will be ignored
    subset : ndarray((n)), optional, default = None
        array of states to be indexed. By default all states in dtrajs will be used

    Returns
    -------
    indexes : list of ndarray( (N_i, 2) )
        For each state, all trajectory and time indexes where this state occurs.
        Each matrix has a number of rows equal to the number of occurances of the corresponding state,
        with rows consisting of a tuple (i, t), where i is the index of the trajectory and t is the time index
        within the trajectory.

    """
    # check input
    dtrajs = _ensure_dtraj_list(dtrajs)
    # select subset unless given
    n = number_of_states(dtrajs)
    if subset is None:
        subset = np.arange(n)
    else:
        if np.max(subset) >= n:
            raise ValueError('Selected subset is not a subset of the states in dtrajs.')
    # histogram states
    hist = count_states(dtrajs, ignore_negative=True)
    # efficient access to which state are accessible
    is_requested = np.ndarray((n), dtype=bool)
    is_requested[:] = False
    is_requested[subset] = True
    # efficient access to requested state indexes
    full2states = np.zeros((n), dtype=int)
    full2states[subset] = range(len(subset))
    # initialize results
    res    = np.ndarray(len(subset), dtype=object)
    counts = np.zeros((len(subset)), dtype=int)
    for i,s in enumerate(subset):
        res[i] = np.zeros((hist[s],2), dtype=int)
    # walk through trajectories and remember requested state indexes
    for i,dtraj in enumerate(dtrajs):
        for t,s in enumerate(dtraj):
            # only index nonnegative state indexes
            if s >= 0 and is_requested[s]:
                k = full2states[s]
                res[k][counts[k],0] = i
                res[k][counts[k],1] = t
                counts[k] += 1
    return res

################################################################################
# sampling from state indexes
################################################################################


def sample_indexes_by_sequence(indexes, sequence):
    """Samples trajectory/time indexes according to the given sequence of states

    Parameters
    ----------
    indexes : list of ndarray( (N_i, 2) )
        For each state, all trajectory and time indexes where this state occurs.
        Each matrix has a number of rows equal to the number of occurrences of the corresponding state,
        with rows consisting of a tuple (i, t), where i is the index of the trajectory and t is the time index
        within the trajectory.
    sequence : array of integers
        A sequence of discrete states. For each state, a trajectory/time index will be sampled at which dtrajs
        have an occurrences of this state

    Returns
    -------
    indexes : ndarray( (N, 2) )
        The sampled index sequence.
        Index array with a number of rows equal to N=len(sequence), with rows consisting of a tuple (i, t),
        where i is the index of the trajectory and t is the time index within the trajectory.

    """
    N = len(sequence)
    res = np.zeros((N,2), dtype=int)
    for t in range(N):
        s = sequence[t]
        i = np.random.randint(indexes[s].shape[0])
        res[t,:] = indexes[s][i,:]

    return res


def sample_indexes_by_state(indexes, nsample, subset=None, replace=True):
    """Samples trajectory/time indexes according to the given sequence of states

    Parameters
    ----------
    indexes : list of ndarray( (N_i, 2) )
        For each state, all trajectory and time indexes where this state occurs.
        Each matrix has a number of rows equal to the number of occurrences of the corresponding state,
        with rows consisting of a tuple (i, t), where i is the index of the trajectory and t is the time index
        within the trajectory.
    nsample : int
        Number of samples per state. If replace = False, the number of returned samples per state could be smaller
        if less than nsample indexes are available for a state.
    subset : ndarray((n)), optional, default = None
        array of states to be indexed. By default all states in dtrajs will be used
    replace : boolean, optional
        Whether the sample is with or without replacement

    Returns
    -------
    indexes : list of ndarray( (N, 2) )
        List of the sampled indices by state.
        Each element is an index array with a number of rows equal to N=len(sequence), with rows consisting of a
        tuple (i, t), where i is the index of the trajectory and t is the time index within the trajectory.

    """
    # how many states in total?
    n = len(indexes)
    # define set of states to work on
    if subset is None:
        subset = np.arange(n)

    # list of states
    res = np.ndarray(len(subset), dtype=object)
    for i, s in enumerate(subset):
        # how many indexes are available?
        m_available = indexes[s].shape[0]
        # do we have no indexes for this state? Then insert empty array.
        if m_available == 0:
            res[i] = np.zeros((0,2), dtype=int)
        elif replace:
            I = np.random.choice(m_available, nsample, replace=True)
            res[i] = indexes[s][I,:]
        else:
            I = np.random.choice(m_available, min(m_available,nsample), replace=False)
            res[i] = indexes[s][I,:]

    return res


def sample_indexes_by_distribution(indexes, distributions, nsample):
    """Samples trajectory/time indexes according to the given probability distributions

    Parameters
    ----------
    indexes : list of ndarray( (N_i, 2) )
        For each state, all trajectory and time indexes where this state occurs.
        Each matrix has a number of rows equal to the number of occurrences of the corresponding state,
        with rows consisting of a tuple (i, t), where i is the index of the trajectory and t is the time index
        within the trajectory.
    distributions : list or array of ndarray ( (n) )
        m distributions over states. Each distribution must be of length n and must sum up to 1.0
    nsample : int
        Number of samples per distribution. If replace = False, the number of returned samples per state could be smaller
        if less than nsample indexes are available for a state.

    Returns
    -------
    indexes : length m list of ndarray( (nsample, 2) )
        List of the sampled indices by distribution.
        Each element is an index array with a number of rows equal to nsample, with rows consisting of a
        tuple (i, t), where i is the index of the trajectory and t is the time index within the trajectory.

    """
    # how many states in total?
    n = len(indexes)
    for dist in distributions:
        if len(dist) != n:
            raise ValueError('Size error: Distributions must all be of length n (number of states).')

    # list of states
    res = np.ndarray(len(distributions), dtype=object)
    for i, dist in enumerate(distributions):
        # sample states by distribution
        sequence = np.random.choice(n, size=nsample, p=dist)
        res[i] = sample_indexes_by_sequence(indexes, sequence)
    #
    return res




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
Created on 15.02.2016

@author: marscher
'''
from itertools import count

import mdtraj
import numpy as np

from pyemma.coordinates.data.featurization._base import Feature
from pyemma.coordinates.data.featurization.util import (_describe_atom,
                                                        cmp_traj)


class CustomFeature(Feature):
    """
    A CustomFeature is the base class for user-defined features. If you want to
    implement a new fancy feature, derive from this class, calculate the quantity
    of interest in the map method and return it as an ndarray.

    If you have defined a map function that should be classed, you don't need to derive a class, but you
    can simply pass a function to the constructor of this class


    Parameters
    ----------
    func : function
        will be invoked with given args and kwargs on mapping traj
    args : list of positional args (optional) passed to func
    kwargs : named arguments (optional) passed to func

    Notes
    -----
    Your passed in function will get a mdtraj.Trajectory object as first argument.

    Examples
    --------
    We define a feature that transforms all coordinates by :math:`x^2`:

    >>> from pyemma.coordinates import source
    >>> from pyemma.datasets import get_bpti_test_data
    >>> inp = get_bpti_test_data()

    Define a function which transforms the coordinates of the trajectory object.
    Note that you need to define the output dimension, which we pass directly in
    the feature construction. The trajectory contains 58 atoms, so the output
    dimension will be 3 * 58 = 174:

    >>> my_feature = CustomFeature(lambda x: (x.xyz**2).reshape(-1, 174), dim=174)
    >>> reader = source(inp['trajs'][0], top=inp['top'])

    pass the feature to the featurizer and transform the data

    >>> reader.featurizer.add_custom_feature(my_feature)
    >>> data = reader.get_output()

    """
    _id = count(0)
    __serialize_version = 0
    __serialize_fields = ('_desc',)

    def __init__(self, fun, dim, description=None, fun_args=(), fun_kwargs=None):
        # we set topology to None here
        self.top = None
        if fun_kwargs is None:
            fun_kwargs = {}
        self._fun = fun
        self._args = fun_args
        self._kwargs = fun_kwargs
        self._dim = dim
        desc = description
        if isinstance(desc, str):
            desc = [desc]
        self.id = next(CustomFeature._id)
        if not desc:
            arg_str = "{args}, {kw}" if self._kwargs else "{args}"
            desc = ["CustomFeature[{id}][0] calling {fun} with args {arg_str}".format(
                id=self.id,
                fun=self._fun,
                arg_str=arg_str, args=self._args, kw=self._kwargs)]
            #if self.dimension > 1:
            #    desc.extend(('CustomFeature[{id}][{i}]'.format(id=self.id, i=i) for i in range(1, self.dimension)))
        elif desc and not (len(desc) == self._dim or len(desc) == 1):
            raise ValueError("to avoid confusion, ensure the lengths of 'description' "
                             "list matches dimension - or give a single element which will be repeated."
                             "Input was {}".format(desc))

        if len(desc) == 1:
            desc *= self.dimension

        self._desc = desc

    def describe(self):
        return self._desc

    def transform(self, traj):
        feature = self._fun(traj, *self._args, **self._kwargs)
        if not isinstance(feature, np.ndarray):
            raise ValueError("your function should return a NumPy array!")
        return feature

    def __eq__(self, other):
        eq = super(CustomFeature, self).__eq__(other)
        if not eq or not isinstance(other, CustomFeature):
            return False
        return self._fun == other._fun and self._args == other._args and self._kwargs == other._kwargs

    def __getstate__(self):
        import warnings
        warnings.warn('We can not save custom functions by now and probably never will. '
                      'Please re-add your custom function after you have restored your Featurizer.')
        return super(CustomFeature, self).__getstate__()

    def __setstate__(self, state):
        super(CustomFeature, self).__setstate__(state)

        def _warn(_):
            raise NotImplementedError('Please re-add your custom feature again! Description was: {}\n'
                                      '>>> featurizer.remove_all_custom_funcs()\n'
                                      '>>> featurizer.add_custom_func(...)'
                                      .format(self.describe()[:30]))
        self._fun = _warn
        self._args = ()
        self._kwargs = {}


class SelectionFeature(Feature):

    """
    Just provide the cartesian coordinates of a selection of atoms (could be simply all atoms).
    The coordinates are flattened as follows: [x1, y1, z1, x2, y2, z2, ...]

    """
    __serialize_version = 0
    __serialize_fields = ('indexes',)
    prefix_label = "ATOM:"

    def __init__(self, top, indexes):
        self.top = top
        self.indexes = np.array(indexes)
        if len(self.indexes) == 0:
            raise ValueError("empty indices")
        self.dimension = 3 * len(indexes)

    def describe(self):
        labels = []
        for i in self.indexes:
            labels.append("%s%s x" %
                          (self.prefix_label, _describe_atom(self.top, i)))
            labels.append("%s%s y" %
                          (self.prefix_label, _describe_atom(self.top, i)))
            labels.append("%s%s z" %
                          (self.prefix_label, _describe_atom(self.top, i)))
        return labels

    def transform(self, traj):
        newshape = (traj.xyz.shape[0], 3 * self.indexes.shape[0])
        return np.reshape(traj.xyz[:, self.indexes, :], newshape)

    def __eq__(self, other):
        eq = super(SelectionFeature, self).__eq__(other)
        if not eq or not isinstance(other, SelectionFeature):
            return False
        return np.all(self.indexes == other.indexes)


class MinRmsdFeature(Feature):

    __serialize_version = 0
    __serialize_fields = ('ref', 'ref_frame', 'name', 'precentered', 'atom_indices',)

    def __init__(self, ref, ref_frame=0, atom_indices=None, topology=None, precentered=False):
        self.top = topology

        assert isinstance(
            ref_frame, int), "ref_frame has to be of type integer, and not %s" % type(ref_frame)

        # Types of inputs
        # 1. Filename+top
        if isinstance(ref, str):
            # Store the filename
            self.name = ref[:]
            ref = mdtraj.load_frame(ref, ref_frame, top=topology)
            # mdtraj is pretty good handling exceptions, we're not checking for
            # types or anything here

        # 2. md.Trajectory object
        elif isinstance(ref, mdtraj.Trajectory):
            self.name = ref.__repr__()[:]
        else:
            raise TypeError("input reference has to be either a filename or "
                            "a mdtraj.Trajectory object, and not of %s" % type(ref))

        self.ref = ref
        self.ref_frame = ref_frame
        self.atom_indices = atom_indices
        self.precentered = precentered
        self.dimension = 1

    def describe(self):
        label = "minrmsd to frame %u of %s" % (self.ref_frame, self.name)
        if self.precentered:
            label += ', precentered=True'
        if self.atom_indices is not None:
            label += ', subset of atoms  '
        return [label]

    def transform(self, traj):
        return np.array(mdtraj.rmsd(traj, self.ref, atom_indices=self.atom_indices), ndmin=2).T

    def __eq__(self, other):
        if not isinstance(other, MinRmsdFeature):
            return False
        eq = super(MinRmsdFeature, self).__eq__(other)
        return (eq and cmp_traj(self.ref, other.ref)
                and self.ref_frame == other.ref_frame
                and np.all(self.atom_indices == other.atom_indices)
                and self.precentered == other.precentered
                )


class AlignFeature(SelectionFeature):
    __serialize_version = 0
    __serialize_fields = ('ref', 'atom_indices', 'ref_atom_indices', 'in_place')

    prefix_label = 'aligned ATOM:'

    def __init__(self, reference, indexes, atom_indices=None, ref_atom_indices=None, in_place=True):
        super(AlignFeature, self).__init__(top=reference.topology, indexes=indexes)
        self.ref = reference
        self.atom_indices = atom_indices
        self.ref_atom_indices = ref_atom_indices
        self.in_place = in_place

    def __eq__(self, other):
        if not isinstance(other, AlignFeature):
            return False
        return (cmp_traj(self.ref, other.ref)
                and np.all(self.atom_indices == other.atom_indices)
                and self.in_place == other.in_place)

    def transform(self, traj):
        if not self.in_place:
            traj = traj.slice(slice(None), copy=True)
        aligned = traj.superpose(reference=self.ref, atom_indices=self.atom_indices,
                                 ref_atom_indices=self.ref_atom_indices)
        # apply selection
        return super(AlignFeature, self).transform(aligned)


class GroupCOMFeature(Feature):
    __serialize_version = 0
    __serialize_fields = ('ref_geom', 'image_molecules', 'group_definitions', 'atom_masses',
                          'masses_in_groups', '_describe')

    def __init__(self, topology, group_definitions, ref_geom=None, image_molecules=False, mass_weighted=True):
        if not (ref_geom is None or isinstance(ref_geom, mdtraj.Trajectory)):
            raise ValueError("argument ref_geom has to be either None or and mdtraj.Trajectory,"
                             " got instead %s" % type(ref_geom))

        self.ref_geom = ref_geom
        self.top = topology
        self.image_molecules = image_molecules
        self.group_definitions = [np.asarray(gf) for gf in group_definitions]
        self.atom_masses = np.array([aa.element.mass for aa in topology.atoms])

        if mass_weighted:
            self.masses_in_groups = [self.atom_masses[aa_in_rr] for aa_in_rr in self.group_definitions]
        else:
            self.masses_in_groups = [np.ones_like(aa_in_rr) for aa_in_rr in self.group_definitions]

        # Prepare and store the description
        self._describe = []
        for group in self.group_definitions:
            for coor in 'xyz':
                self._describe.append('COM-%s of atom group [%s..%s] '%(coor, group[:3], group[-3:]))
                # TODO consider including the ref_geom and image_molecule arsg here?
        self.dimension = 3 * len(self.group_definitions)

    def describe(self):
        return self._describe

    def transform(self, traj):
        # TODO: is it possible to avoid copy? Otherwise the trajectory is altered...
        traj_copy = traj[:]
        COM_xyz = []
        if self.ref_geom is not None:
            traj_copy = traj_copy.superpose(self.ref_geom)
        if self.image_molecules:
            traj_copy = traj_copy.image_molecules()
        for aas, mms in zip(self.group_definitions, self.masses_in_groups):
            COM_xyz.append(np.average(traj_copy.xyz[:, aas, ], axis=1, weights=mms))
        return np.hstack(COM_xyz)

    def __eq__(self, other):
        eq = super(GroupCOMFeature, self).__eq__(other)
        if not eq or not isinstance(other, GroupCOMFeature):
            return False
        return (cmp_traj(self.ref_geom, other.ref_geom) and self.image_molecules == other.image_molecules
                and all(np.array_equal(g1, g2) for g1, g2 in zip(self.group_definitions, other.group_definitions))
                and all(np.array_equal(m1, m2) for m1, m2 in zip(self.masses_in_groups, other.masses_in_groups))
        )


class ResidueCOMFeature(GroupCOMFeature):

    __serialize_version = 0
    __serialize_fields = ('residue_indices', 'scheme')

    def __init__(self, topology, residue_indices, residue_atoms, scheme, ref_geom=None, image_molecules=False,
                 mass_weighted=True):
        super(ResidueCOMFeature, self).__init__(topology, residue_atoms, mass_weighted=mass_weighted, ref_geom=ref_geom,
                                                image_molecules=image_molecules)

        # This are the only extra attributes that residueCOMFeature should have
        self.residue_indices = residue_indices
        self.scheme = scheme

        # Overwrite the self._describe attribute, this way the method of the superclass can be used "as is"
        self._describe = []
        for ri in self.residue_indices:
            for coor in 'xyz':
                self._describe.append('%s COM-%s (%s)' % (topology.residue(ri), coor, self.scheme))
                # TODO consider including the ref_geom and image_molecule arsg here?

    def __eq__(self, other):
        eq = super(ResidueCOMFeature, self).__eq__(other)
        if not eq or not isinstance(other, ResidueCOMFeature):
            return False
        return np.all(self.residue_indices == other.residue_indices) and self.scheme == other.scheme



# This file is part of PyEMMA.
#
# Copyright (c) 2017 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as _np
from pyemma.coordinates import assign_to_centers as _assign_to_centers

__all__ = [
    'DoubleWellSampler']

class DoubleWellSampler(object):
    '''Continuous multi-ensemble MCMC process in an asymmetric double well potential'''
    def __init__(self, random_seed=None):
        self.xmin = -1.8
        self.xmid = 0.127
        self.xmax = 1.7
        self.step = 0.6
        self.nstates = 50
        self.stride = None
        edges = _np.linspace(self.xmin, self.xmax, self.nstates + 1)
        self.x = 0.5 * (edges[1:] + edges[:-1])
        epot = self.potential(self.x)
        self.pi = _np.exp(-epot)
        self.pi[:] = self.pi / self.pi.sum()
        self.f = -_np.log(self.pi)

        if random_seed is not None:
            self._rng = _np.random.RandomState(random_seed)
        else:
            self._rng = _np.random
        self.random_seed = random_seed

    @property
    def centers(self):
        return self.x.reshape(-1, 1)

    def _potential(self, x):
        try:
            return _np.asarray(tuple(map(self._potential, x)))
        except TypeError:
            if x < self.xmin or x > self.xmax:
                return _np.inf
            return x * (0.5 + x * (x * x - 2.0))

    def _bias(self, x, kbias, xbias):
        try:
            return 0.5 * kbias * (_np.asarray(x) - xbias)**2
        except TypeError:
            return 0.0

    def potential(self, x, kt=1.0, kbias=None, xbias=None):
        return (self._potential(x) + self._bias(x, kbias, xbias)) / kt

    def mcmc(self, xinit, length, kt=1.0, kbias=None, xbias=None, stride=None):
        xtraj = _np.zeros(shape=(length + 1,), dtype=_np.float64)
        etraj = _np.zeros(shape=(length + 1,), dtype=_np.float64)
        xtraj[0] = xinit
        etraj[0] = self.potential(xinit, kt=kt, kbias=kbias, xbias=xbias)
        for i in range(length):
            x_candidate = xtraj[i] + self.step * (self._rng.rand() - 0.5)
            e_candidate = self.potential(x_candidate, kt=kt, kbias=kbias, xbias=xbias)
            if e_candidate < etraj[i] or self._rng.rand() < _np.exp(etraj[i] - e_candidate):
                xtraj[i + 1] = x_candidate
                etraj[i + 1] = e_candidate
            else:
                xtraj[i + 1] = xtraj[i]
                etraj[i + 1] = etraj[i]
        if stride is not None and stride > 1:
            xtraj, etraj = _np.ascontiguousarray(xtraj[::stride]), _np.ascontiguousarray(etraj[::stride])
        return xtraj, etraj

    def _draw(self, xinit=None, right=False, weighted=True):
        if xinit is None:
            if right:
                pad = 0.2
                return self._rng.rand() * (self.xmax - self.xmid - 2.0 * pad) + self.xmid + pad
            return self._rng.choice(self.x, size=1, p=self.pi if weighted is True else None)
        return xinit

    def sample(self, ntraj=1, xinit=None, length=10000):
        trajs = [self.mcmc(
             self._draw(xinit), length=length, stride=self.stride)[0] for i in range(ntraj)]
        return dict(
            trajs=trajs,
            dtrajs=_assign_to_centers(trajs, centers=self.centers))

    def us_sample(self, ntherm=11, us_fc=20.0, us_length=500, md_length=1000, nmd=20):
        xbias = _np.linspace(self.xmin + 0.2, self.xmax - 0.2, ntherm + 1)
        xbias = (0.5 * (xbias[1:] + xbias[:-1])).tolist()
        kbias = [us_fc] * len(xbias)
        us_trajs = [self.mcmc(
            x, us_length, kbias=k, xbias=x, stride=self.stride)[0] for k, x in zip(kbias, xbias)]
        md_trajs = [self.mcmc(
            self._draw(right=True), md_length, stride=self.stride)[0] for i in range(nmd)]
        dtrajs = _assign_to_centers(us_trajs + md_trajs, centers=self.centers)
        return dict(
            us_trajs=us_trajs,
            us_dtrajs=dtrajs[:ntherm],
            us_centers=xbias,
            us_force_constants=kbias,
            md_trajs=md_trajs,
            md_dtrajs=dtrajs[ntherm:])

    def mt_sample(self, kt0=1.0, kt1=5.0, length0=10000, length1=10000, n0=10, n1=10):
        trajs = []
        utrajs = []
        ttrajs = []
        for i in range(n0):
            x, u = self.mcmc(
                self._draw(right=True), length0, kt=kt0, stride=self.stride)
            trajs.append(x)
            utrajs.append(u)
            ttrajs.append(_np.asarray([kt0] * x.shape[0]))
        for i in range(n1):
            x, u = self.mcmc(
                self._draw(weighted=False), length1, kt=kt1, stride=self.stride)
            trajs.append(x)
            utrajs.append(u)
            ttrajs.append(_np.asarray([kt1] * x.shape[0]))
        return dict(
            trajs=trajs,
            energy_trajs=utrajs,
            temp_trajs=ttrajs,
            dtrajs=_assign_to_centers(trajs, centers=self.centers))
